import sys, time, traceback, types, inspect
from collections.abc import Iterable

from .client import JellyClient


global __jelly__
__jelly__ = None
def init (*path, **config):
  global __jelly__
  route = []
  for p in path:
    if not isinstance(p, str):
      raise ValueError('path arguments must be string')
    for r in p.split('/'):
      if r:
        route.append(r)
  
  for i, item in enumerate(route):
    if i == 0:
      config['workspace'] = item
    elif i == 1:
      config['project'] = item
    elif i == 2:
      config['process'] = item

  if not __jelly__:
    __jelly__ = JellyClient()
    __jelly__.activate(**config)
  return __jelly__

def serve ():
  global __jelly__
  __jelly__.serve()

def is_primitive (value):
  primtypes = ['int', 'str', 'bool', 'float', 'long', 'NoneType']
  return type(value).__name__ in primtypes

def is_iterable (value):
  return isinstance(value, Iterable)

def is_list (value):
  return isinstance(value, list) or isinstance(value, tuple)

def is_dict (value):
  return isinstance(value, dict)

def is_module (value):
  return type(value).__name__ == 'module'

def is_builtin (value):
  return is_primitive(value) or is_iterable(value) or is_dict(value) or is_module(value)

def is_instance (value):
  return not is_builtin(value)

def flatten (alist, result=None):
  if result is None:
    result = []
  for item in alist:
    if is_list(item):
      flatten(item, result)
    elif is_primitive(item):
      result.append(item)
    else:
      raise Exception('cannot flatten type: ' + type(item).__name__)
  return result


from .html import Elements, ComponentElement

elements = Elements(ComponentElement(None))
div = elements.div
span = elements.span
button = elements.button
input_ = elements.input


import ssl, json, time, sys, queue, atexit, logging
import paho.mqtt.client as paho


LOG = logging.getLogger(__name__)

class MQTTJSONMessage:
  def __init__ (self, topic, payload):
    self.topic = topic
    self.ts = time.time() * 1000.0
    self.payload = json.loads(payload)

class MQTTSubscriber:
  def __init__ (self, mqtt, *topics):
    self.mqtt = mqtt
    self.topics = topics
    self.queue = queue.Queue()

  def add_message (self, message):
    self.queue.put(message)
  
  def is_subscribed (self, topic):
    return True # TODO: check for method

  def __iter__ (self):
    while not self.queue.empty():
      msg = self.queue.get()
      yield MQTTJSONMessage(msg.topic, msg.payload)

  def consume (self, sleep=0.01):
    while 1:
      for msg in self:
        yield msg
      time.sleep(sleep)


class MQTT:
  def __init__ (self, host, port, client_id, use_ssl=True, transport='websockets'):
    self.use_ssl = use_ssl
    self.client_id = client_id
    self.mqtt_host = host
    self.mqtt_port = port
    self.subscriptions = []
    self.is_connected = None
    self.transport = transport

    self.client = paho.Client(transport=self.transport, client_id=self.client_id)
    self.client.on_message = self.on_message
    self.client.on_connect = self.on_connect
    self.client.on_disconnect = self.on_disconnect
    self.client.max_inflight_messages_set(0)

    if self.use_ssl:
        ssl_ctx = ssl.create_default_context()
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE
        self.client.tls_set_context(ssl_ctx)
        #self.client.tls_insecure_set(True)

    atexit.register(self.wait_for_publish)

  def subscribe (self, *topics):
    sub = MQTTSubscriber(self, *topics)
    self.subscriptions.append(sub)
    return sub

  def on_connect (self, *args, **kw):
    sys.stdout.write('connected: %s\n' % self.client_id)
    self.is_connected = True
    self.client.subscribe([('#', 1,)])
    
  def on_disconnect (self, *args, **kw):
    sys.stdout.write('disconnected\n')
    LOG.warning('mqtt disconnected')
    self.is_connected = False
    
  def on_message (self, client, _, msg):
    sys.stdout.write('received: %s\n' % msg.topic)
    for sub in self.subscriptions:
      if not sub.is_subscribed(msg.topic):
        continue
      sub.add_message(msg)
    
  def publish (self, topic, msg):
    info = self.client.publish(topic, json.dumps(msg))
    return info

  def wait_for_publish (self):
    # TODO: save unsent messages to file until confirmed
    time.sleep(1) 

  def connect (self):
    result = self.client.connect(self.mqtt_host, self.mqtt_port, keepalive=10)
    self.client.loop_start()


import time, sys, threading, logging, traceback, uuid, os, builtins
from collections import defaultdict, OrderedDict
from uuid import uuid4, getnode
from threading import Thread
from queue import Queue

import multiprocessing
ENV = multiprocessing.get_context("spawn")

from .transport import MQTT
from .logger import JellyLogger, LoggingLevel, _print

LOG = logging.getLogger(__name__)


class JellyRuntime:
  def __init__ (self, client, scope):
    self.client = client
    self.scope = scope

  def evaluate (self, code):
    LOG.log(LoggingLevel.STATUS, 'starting evaluate')
    try:
      exec(code, globals(), self.scope)
    except Exception as err:
      #LOG.error("%s", err)
      LOG.exception(err)
    finally:
      LOG.log(LoggingLevel.STATUS, 'evaluate complete')

class JellyClient:
  def __init__ (self, workspace=None, project=None, process=None, version='v1', host='ws.web-jelly.com', port=1884):
    self.version = version
    self.client_id = str(uuid.uuid4())    
    self.config = dict(
      workspace=workspace, project=project, process=process
    )

    self.transport = MQTT(host, port, self.client_id)
    self.transport.connect()

    self.logging = JellyLogger(self)
    self.rpc = JellyRPC(self)
    self.runtimes = {}
    self.api = dict()

  def set_app (self, api):
    if hasattr(api, '__dict__'):
      self.api = api.__dict__
    elif isinstance(api, dict):
      self.api = api
    else:
      raise ValueError('api must be a dict or object')

  def serve (self):
    while 1:
      time.sleep(0.1)

  def evaluate (self, id, code):
    if not id in self.runtimes:
      self.runtimes[id] = JellyRuntime(self, self.api)
    self.runtimes[id].evaluate(code)

  def kill (self):
    LOG.log(LoggingLevel.STATUS, 'process stopping')
    self.transport.wait_for_publish()
    os._exit(0)
  
  def introduce (self):
    self.publish('status', self.config)

  def make_topic (self, method):
    c = self.config
    topic = "%s/%s/%s/%s/%s/%s/%s/%s" % (
      self.version, c['workspace'], 'client', self.client_id, 
      'outbox', method, c['project'], c['process']
    )
    return topic

  def activate (self, **kw):
    self.config.update(**kw)
    app_path = os.path.abspath(self.config.get('app', '.'))
    sys.path.append(app_path)
    self.logging.activate(**self.config.get('logging', {}))
    self.rpc.activate(**self.config.get('rpc', {}))
    self.introduce()
    LOG.log(LoggingLevel.STATUS, 'process started')

  def publish (self, method, payload):
    topic = self.make_topic(method)
    payload.update(scope=self.get_scope())
    self.transport.publish(topic, payload)

  def get_scope (self):
    c = self.config
    return "%s/%s/%s" % (c['workspace'], c['project'], c['process'])  


# == RPC ==
class Action:
  def __init__ (self, client, action_id, handler):
    self.client = client
    self.id = action_id
    self.handler = handler
  
  def execute (self, param):
    self.client.execute_action(self.id, self.handler, param)

class JellyRPC:
  def __init__ (self, client):
    self.client = client
    self.transport = client.transport
    self.active_calls = {}
    self.actions = {}
    self._main_thread = None
    
  def activate (self, **kwargs):
    self._main_thread = Thread(target=self._main_loop)
    self._main_thread.daemon = True
    self._main_thread.start()

  def _main_loop (self):
    requests = self.transport.subscribe()
    while 1:
      for req in requests:
        try:
          self._handle_request(req)
        except Exception as err:
          traceback.print_exc()
          LOG.exception(err)
      time.sleep(0.01)

  def _handle_request (self, req):
    _print('handle request', req.payload)
    type = req.payload.get('type').lower()
    args = req.payload.get('payload')
    #LOG.info("> %s(%s)", type, args)
    if type == 'evaluate':
      self.client.evaluate(args['id'], args['code'])
    elif type == 'connect':
      self.client.introduce()
    elif type == 'kill':
      self.client.kill()




import inspect, time, sys, builtins, logging, traceback
from uuid import uuid4

_print = builtins.print

def now ():
  return int(time.time() * 1000)

def uuid ():
  return str(uuid4())

def function_name (fn):
  klass = fn.__self__.__class__.__name__
  return "%s.%s" % (klass, fn.__name__)

def _parse_messages (list_of_messages):
  return ' '.join(map(str, list_of_messages))

class LoggingLevel:
  STATUS = 60
  CRITICAL = 50
  ERROR = 40
  WARNING = 30
  INFO = 20
  DEBUG = 10
  NOTSET = 0

def log_level_name (level):
  if level == LoggingLevel.STATUS:
    return 'STATUS'
  elif level == LoggingLevel.CRITICAL:
    return 'CRITICAL'
  elif level == LoggingLevel.ERROR:
    return 'ERROR'
  elif level == LoggingLevel.WARNING:
    return 'WARNING'
  elif level == LoggingLevel.INFO:
    return 'INFO'
  elif level == LoggingLevel.NOTSET:
    return 'NOTSET'
  return ''


class JellyfishLoggingHandler(logging.NullHandler):
  def __init__ (self, client):
    super(JellyfishLoggingHandler, self).__init__()
    self.client = client
    
  def emit(self, record):
    publish_logging_record(self.client, record)


def publish_logging_record (client, record):
  record.levelname = log_level_name(record.levelno)
  message = record.getMessage()
  #message = record.msg
  if record.exc_info:
    try:
      exc = traceback.format_exception(*record.exc_info)
      message += '\n' + '\n'.join(exc)
    except:
      _print('WARNING: failed to format exception')
  #if record.exc_text:
  #  print('publish record exception', record)
  #  message += '\n' + record.exc_text
  #print(record.__dict__)
  #if isinstance(record.msg, str):
  #  message = record.getMessage()
  #else:
  #message = record.msg
  try:
    args = record.args
    if args and not isinstance(args, (tuple, list,)):
      args = [args]  
    items = [client.printer.print_any(arg) for arg in args]
    
    client.publish_log(dict(
      name=record.name, message=record.msg, level=record.levelname,
      levelno=record.levelno, ts=time.time()*1000.0,
      threadid=record.thread, threadName=record.threadName,
      processid=record.process, module=record.module,
      lineno=record.lineno, pathname=record.pathname,
      args=items
    ))
  except Exception as err:
    _print(err)
    #_print(traceback.format_exception(err))
    #traceback.print_exc()

class JellyLogger:

  def __init__ (self, client):
    self.client = client
    self.active_request = None
    self.level = LoggingLevel.INFO
    self.current_print_line = None
    self.printer = ObjectPrinter()

  def activate (self, level='INFO'):
    # NOTE:
    #   for some reason, adding a handler messes with basicConfig somehow
    #   have to use logging factory to intercept logs 
    self.level = LoggingLevel.DEBUG
    _print('activated logging with level', self.level)
    old_factory = logging.getLogRecordFactory()
    def record_factory(*args, **kwargs):
      record = old_factory(*args, **kwargs)
      publish_logging_record(self, record)
      return record
    logging.setLogRecordFactory(record_factory)
    
    print_logger = logging.getLogger('print')
    def new_print (*args, **kwargs):
      end = kwargs.get('end')
      log_msg = " ".join(["%s" for _ in range(len(args))])
      print_logger.info(log_msg, *args)
      _print(*args, **kwargs)
    builtins.print = new_print    
    
    # NOTE: do not do this: (test with basicConfig)
    # logging.root.addHandler(JellyfishLoggingHandler(self))
    
  def publish_log (self, log):
    level = log.get('levelno') # logging.getLevelName(self.level)
    log_level = self.level # logging.getLevelName(log.get('level'))
    #_print('LOG', level, log_level)
    if not isinstance(level, int) or not isinstance(log_level, int):
      return _print('skipped log', level, log_level)
    if log_level <= level:
      log['_id'] = str(uuid())
      #_print('PUBLISH')
      self.client.publish('log', log)
    else:
      _print('skipped log', level, log_level)

  def set_active_request (self, req_id):
    self.active_request = req_id

  



class ObjectPrinter:
  def __init__ (self, full=False):
    self.full = full

  def print_any (self, value):
    result = {'type': 'unknown', 'value': None}
    t = type(value).__name__
    if value is None:
      result = dict(type='null')
    elif isinstance(value, Exception):
      result = {'type': 'error', 'value': dict(type=t, message=str(value))}
    elif t == 'int' or t == 'str' or t == 'bool' or t == 'long' or t == 'float' or t == 'NoneType':
      result = {'type': t, 'value': value}
    elif t == 'dict':
      #result = {'type': 'dict', 'value': {}}
      result = self.print_dict(value)
    elif t == 'list':# or isinstance(value, Iterable):
      #result = {'type': 'list', 'value': []}
      result = self.print_list(value)
    elif t == 'function':
      result = {'type': 'function', 'value': value.__name__}
    elif t == 'type':
      result = {'type': 'type', 'value': value.__name__}
    elif t == 'module':
      result = {'type': 'module', 'value': value.__name__ + '.py'}
    elif hasattr(value, '__render__'):
      return self.print_any(value.__render__())
    elif hasattr(value, '__serialize__'):
      result = value.__serialize__(self.print_any)
    elif hasattr(value, '__dict__'):
      result = self.print_instance(value)
    else:
      result = self.print_instance(value)
    return result

  def print_dict (self, d):
    result = {'type': 'dict', 'value': {}}
    for k,v in d.items():
      result['value'][k] = self.print_any(v)
    return result
    
  def print_list (self, l):
    result = []
    for val in l:
      result.append(self.print_any(val))
    return {'type': 'list', 'value': result}

  def print_instance (self, t):
    result = {}
    #if hasattr(t, '__dict__'):
    #  for k, v in t.__dict__.items():
    #    if k.startswith('_'):
    #      continue
    #    result[k] = self.print_any(v)
    #else:
    #  return {'type': 'str', 'value': str(t)}
    #for k, v in t.__class__.__dict__.items():
    #  if k.startswith('_'):
    #    continue
    #  result[k] = self.print_any(v)
    result = self.print_dict(result)
    return {'type': 'object', 'class': t.__class__.__name__, 'value': result}

    





"""

class Logger:
  def __init__ (self, paths=None):
    self.paths = paths # files not in here will not get logged
    self.logs = []
    
  def clear (self):
    self.logs = []

  def log (self, *messages):
    return self._create_log('LOG', inspect.stack(), *messages)

  def error (self, *messages):
    return self._create_log('ERROR', inspect.stack(), *messages)

  def exception (self, exc):
    type, exc, tb = sys.exc_info()
    info = inspect.getinnerframes(tb)
    info.reverse()
    return self._create_log('ERROR', info, type.__name__, str(exc))

  def parse_value (self, value):
    if is_primitive(value):
      return value
    return repr(value)

  def is_app_frame (self, frame):
    # TODO: use self.paths
    return True
    
  def get_object_path (self, value):
    path = []
    if hasattr(value, '__module__'):
      path.append(getattr(value, '__module__'))
    if hasattr(value, '__class__'):
      path.append(getattr(value, '__class__').__name__)
    return path

  def parse_locals (self, fn_name, _locals, _globals):
    result = {}
    is_method = 'self' in _locals

    parent = []
    fn_object = None
    #if is_method:
    #  parent = self.get_object_path(_locals['self']) + [fn_name]
    #  fn_object = getattr(_locals['self'], fn_name) #.__dict__[fn_name]
    #elif fn_name in _globals:
    #  parent = [_globals['__name__'], fn_name]
    #  fn_object = _globals[fn_name]
    
    for k, v in _locals.items():
      if k.startswith('__') or k == 'self':
        continue
      result[k] = self.parse_value(v)
    return {'locals': result, 'fullpath': '.'.join(parent)}

  def _create_log (self, log_type, stack, *messages):
    items = []
    for msg in messages:
      item = {
        'value': msg, # self.parse_value(msg), 
        'object': id(msg), 'type': type(msg).__name__
      }
      items.append(item)
    
    sdata = []
    for frame in stack:
      if not self.is_app_frame(frame):
        continue
      ldata = self.parse_locals(frame.function, frame.frame.f_locals, frame.frame.f_globals)
      frame_data = {
        #'locals': ldata['locals'], 
        'fullpath': ldata['fullpath'],
        'line': frame.lineno, 'file': frame.filename, 'function': frame.function
      }
      sdata.append(frame_data)

    log = {
      'type': log_type, 'time': int(time.time() * 1000.0),
      'items': items,
      'stack': sdata
    }
    self.logs.append(log)
    return log


class PrintObserver:
  def __init__ (self):
    #self.command = command
    #self.service = service
    self.buffer = [[]]
    self.real_print = builtins.print

  def is_empty (self):
    return len(self.buffer) == 0

  def get (self):
    return self.buffer.pop(0)

  def digest (self):
    while not self.is_empty():
      yield self.get()
      
  def stop (self):
    builtins.print = self.real_print

  def start (self, callback):
    def _print (*args, file=None, end=''):
      self.buffer.append([])
      for item in args:
        #sys.stdout.write(str(item) + ' ')
        self.write(item)
      #sys.stdout.write(end or '\n')
      callback(self.buffer[-1])
    builtins.print = _print

  def write (self, item):
    line = self.buffer[-1]
    line.append(item)
    #if type(item) is str and item.endswith('\n'):
      #self.service.emit(self.command, 'LOG', self.buffer)
    #  self.buffer.append([])

  def flush (self):
    pass

"""